import importlib.resources
import json
import logging
from html import escape

import pandas as pd
from IPython.display import HTML, display

import framedisplay

LOG = logging.getLogger(__name__)

from .__version__ import __version__

JS_FILEPATH = str(importlib.resources.files(framedisplay).joinpath("js", "framedisplay.min.js"))
JS_CDN_URL = f"https://cdn.jsdelivr.net/gh/nsarang/framedisplay@v{__version__}/framedisplay/js/framedisplay.min.js"


def initialize():
    """
    Initialize the FrameDisplay JavaScript in Jupyter Notebook.
    This is optional and is only needed for offline usage.
    """
    with open(JS_FILEPATH, "r", encoding="utf-8") as f:
        js_content = f.read()
    display(HTML(f'<script type="text/javascript">{js_content}</script>'))
    LOG.info("FrameDisplay JavaScript initialized.")


def configure(config: dict = None, reset: bool = False, return_html: bool = False) -> None:
    """
    Configure FrameDisplay settings globally.

    Parameters
    ----------
    config : dict
        Configuration options for FrameDisplay. These will be passed to the JavaScript
        as a global variable `FrameDisplayConfig`. See the documentation for available options.
    reset : bool, optional
        If True, resets the FrameDisplayConfig overrides to empty before applying the new config.
        This is useful if you want to ensure that only the provided config options are used,
        without any previously set options interfering. Defaults to False.
    return_html : bool, optional
        If True, return the HTML string instead of displaying it. Defaults to False.
    """
    if config is None and not reset:
        raise ValueError("Either `config` must be provided or `reset` must be True.")

    config_script = ""
    if reset:
        config_script += """
            <script>
                window.FrameDisplayConfig = {};
            </script>
        """
    config_json = json.dumps(config)
    config_script += f"""
        <script>
            window.FrameDisplayConfig = Object.assign(window.FrameDisplayConfig || {{}}, {config_json});
        </script>
    """
    if return_html:
        return config_script
    display(HTML(config_script))


def get_type(series: pd.Series) -> str:
    """
    Get a simplified type name from a pandas dtype.
    """
    dtype = series.dtype if hasattr(series, "dtype") else series
    if pd.api.types.is_integer_dtype(dtype):
        return "int"
    elif pd.api.types.is_float_dtype(dtype):
        return "float"
    elif pd.api.types.is_string_dtype(dtype):
        return "string"
    elif pd.api.types.is_datetime64_any_dtype(dtype):
        return "datetime"
    elif pd.api.types.is_bool_dtype(dtype):
        return "bool"
    elif isinstance(dtype, pd.CategoricalDtype):
        return "category"
    else:
        return "object"


def dataframe_to_html(df: pd.DataFrame) -> str:
    """
    Minimal HTML generator for displaying a pandas DataFrame.
    """

    # Header columns
    dtypes = df.convert_dtypes().apply(get_type).values
    header_cols = "".join(
        f"<th data-dtype={ctype}>{escape(str(col))}</th>" for col, ctype in zip(df.columns, dtypes)
    )

    # Body rows
    rows = []
    for idx, row in df.iterrows():
        cells = [f"<th>{escape(str(idx))}</th>"]  # Index cell
        for value in row:
            if pd.isna(value):
                cells.append('<td><code class="null-cell">null</code></td>')
            else:
                cells.append(f"<td>{escape(str(value))}</td>")

        rows.append(f"<tr>{''.join(cells)}</tr>")

    return f"""
        <table border="1" class="frame-display-table">
            <thead>
                <tr style="text-align: right;">
                    <th></th> <!-- Index column -->
                    {header_cols}
                </tr>
            </thead>
            <tbody>
                {"".join(rows)}
            </tbody>
        </table>
    """


def frame_display(
    df: pd.DataFrame,
    jspath: str = None,
    embed_style: str = None,
    return_html: bool = False,
    config: dict = None,
    reset_config: bool = False,
) -> None:
    """
    Display a DataFrame as HTML in Jupyter Notebook.

    Parameters
    ----------
    df : pd.DataFrame
        The DataFrame to display.
    jspath : str, optional
        The path to the FrameDisplay JavaScript file. Defaults to the CDN URL.
    embed_style : str, optional
        Method for including JavaScript and CSS resources:
        - None: Reference JavaScript from the external source specified by `jspath` (default)
        - 'css_only': Embed only CSS styles inline. This is useful for email clients that don't support external scripts.
        - 'all': Embed both JavaScript and CSS inline
    return_html : bool, optional
        If True, return the HTML string instead of displaying it. Defaults to False.
    config : dict, optional
        Configuration options for FrameDisplay. These will be passed to the JavaScript
        as a global variable `FrameDisplayConfig`. See the documentation for available options.
    reset_config : bool, optional
        If True, resets the FrameDisplayConfig overrides to empty before applying the new config.
        This is useful if you want to ensure that only the provided config options are used,
        without any previously set options interfering. Defaults to False.

    Returns
    -------
    str or None
        If return_html is True, returns the HTML string. Otherwise, displays the content
        and returns None.
    """
    if config is not None or reset_config:
        config_script = configure(config or {}, reset=reset_config, return_html=True)
    else:
        config_script = ""

    if embed_style is not None:
        if embed_style == "css_only":
            script_content = f"""
            <style>
                {(importlib.resources.files(framedisplay) / "js/src/styles.css").read_text()}
            </style>
            """
        elif embed_style == "all":
            script_content = f"""
            <script type="text/javascript">
                {(importlib.resources.files(framedisplay) / "js/framedisplay.min.js").read_text()}
            </script>
            """
        else:
            raise ValueError(
                "Invalid value for `embed_style`. Must be 'all' or 'css_only' or None."
            )
    else:
        jspath = jspath or JS_CDN_URL
        script_content = f"<script src='{escape(jspath)}'></script>"

    html_content = f"""
        <div class="table-container">
            {config_script}
            {script_content}
            {dataframe_to_html(df)}
        </div>
    """
    if return_html:
        return html_content
    display(HTML(html_content))


def _get_dataframe_subset(
    df: pd.DataFrame, max_cells: int = None, max_rows: int = None
) -> pd.DataFrame:
    "Helper function to get a subset of the DataFrame based on max_cells or max_rows."
    assert max_cells is not None or max_rows is not None
    if max_cells is not None:
        nrows, ncols = df.shape
        nrows = max(1, max_cells // ncols)
        if max_rows is not None:
            nrows = min(nrows, max_rows)
        return df.head(nrows)
    return df.head(max_rows)


def integrate_with_pandas(max_cells: int = 10000, max_rows: int = None, **kwargs) -> None:
    """
    This function patches the pandas DataFrame class to use FrameDisplay for HTML
    rendering in Jupyter notebooks and other environments that support rich display.
    After calling this function, all DataFrames will automatically use FrameDisplay
    when displayed in notebook cells.

    Parameters
    ----------
    max_cells : int, optional
        Maximum number of cells to display in the DataFrame. If the DataFrame exceeds
        this size, it will be truncated. Defaults to 10000.
    **kwargs : dict
        Additional keyword arguments to pass to `frame_display`.

    Notes
    -----
    This modifies the global pandas DataFrame._repr_html_ method.
    """
    if not hasattr(pd.DataFrame, "_repr_html_original"):
        pd.DataFrame._repr_html_original = pd.DataFrame._repr_html_
    pd.DataFrame._repr_html_ = lambda df: frame_display(
        _get_dataframe_subset(df, max_cells=max_cells, max_rows=max_rows),
        return_html=True,
        **kwargs,
    )



# Requires pytest-mock: pip install pytest-mock
# This provides the 'mocker' fixture as a pytest-compatible alternative to unittest.mock

import numpy as np
import pandas as pd
import pytest
from bs4 import BeautifulSoup
from hypothesis import given
from hypothesis import strategies as st

from framedisplay.display import (
    dataframe_to_html,
    frame_display,
    get_type,
    initialize,
    integrate_with_pandas,
)


@pytest.fixture
def sample_df():
    return pd.DataFrame(
        {"int_col": [1, 2, 3], "str_col": ["a", "b", "c"], "float_col": [1.1, 2.2, np.nan]}
    )


class TestCoreHTMLGeneration:
    """Core functionality and HTML generation."""

    def test_basic_dataframe_structure(self, sample_df):
        html = dataframe_to_html(sample_df)
        soup = BeautifulSoup(html, "html.parser")

        table = soup.find("table", class_="frame-display-table")
        assert table is not None

        headers = soup.find("thead").find_all("th")
        assert len(headers) == 4  # Index + 3 columns
        assert headers[1].text == "int_col"
        assert headers[2].text == "str_col"
        assert headers[3].text == "float_col"

        rows = soup.find("tbody").find_all("tr")
        assert len(rows) == 3

    def test_null_values_handling(self):
        df = pd.DataFrame({"mixed_nulls": [1, None, 3, np.nan, 5]})

        html = dataframe_to_html(df)
        soup = BeautifulSoup(html, "html.parser")

        null_cells = soup.find_all("code", class_="null-cell")
        assert len(null_cells) == 2  # None and np.nan
        for cell in null_cells:
            assert cell.text == "null"

    def test_empty_dataframe(self):
        df = pd.DataFrame()
        html = dataframe_to_html(df)
        soup = BeautifulSoup(html, "html.parser")

        assert soup.find("table", class_="frame-display-table") is not None
        assert soup.find("thead") is not None
        assert soup.find("tbody") is not None


class TestSecurityAndEscaping:
    """XSS prevention and input sanitization."""

    def test_xss_prevention_in_data(self):
        df = pd.DataFrame(
            {"malicious": ['<script>alert("xss")</script>', "<img src=x onerror=alert(1)>"]}
        )

        html = dataframe_to_html(df)
        assert "<script>" not in html
        assert "<img" not in html
        assert "&lt;script&gt;" in html
        assert "&lt;img" in html

    def test_xss_prevention_in_column_names(self):
        df = pd.DataFrame({'<script>alert("xss")</script>': [1, 2, 3]})

        html = dataframe_to_html(df)
        assert "<script>alert(" not in html
        assert "&lt;script&gt;" in html

    def test_malicious_js_path_escaping(self):
        df = pd.DataFrame({"A": [1]})
        malicious_path = '"><script>alert(1)</script>'

        html = frame_display(df, jspath=malicious_path, return_html=True)
        assert "<script>alert(1)" not in html
        assert "&quot;&gt;&lt;script&gt;" in html


class TestDataTypeHandling:
    """Data type detection and handling."""

    @pytest.mark.parametrize(
        "dtype,expected",
        [
            (np.dtype("int64"), "int"),
            (np.dtype("float64"), "float"),
            (pd.StringDtype(), "string"),
            (np.dtype("bool"), "bool"),
            (np.dtype("datetime64[ns]"), "datetime"),
            (pd.CategoricalDtype(["A", "B"]), "category"),
            (np.dtype("object"), "string"),  # pandas considers object as string
            (np.dtype("complex128"), "object"),
        ],
    )
    def test_get_type_function(self, dtype, expected):
        assert get_type(dtype) == expected

    def test_datetime_columns_get_correct_dtype(self):
        df = pd.DataFrame({"dates": pd.date_range("2023-01-01", periods=3)})

        html = frame_display(df, return_html=True)
        assert "data-dtype=datetime" in html

    def test_mixed_data_types_in_output(self):
        df = pd.DataFrame(
            {
                "integers": [1, 2, 3],
                "floats": [1.1, 2.2, np.inf],
                "strings": ["a", "b", "c"],
                "booleans": [True, False, True],
            }
        )

        html = frame_display(df, return_html=True)
        assert "data-dtype=int" in html
        assert "data-dtype=float" in html
        assert "data-dtype=string" in html
        assert "data-dtype=bool" in html


class TestFrameDisplayFunction:
    """Main user-facing function behavior."""

    def test_return_vs_display_modes(self, sample_df, mocker):
        mock_display = mocker.patch("framedisplay.display.display")

        # Default behavior: display in Jupyter
        result = frame_display(sample_df)
        assert result is None
        mock_display.assert_called_once()

        # Return HTML mode
        html = frame_display(sample_df, return_html=True)
        assert isinstance(html, str)
        assert "table-container" in html

    def test_custom_js_path(self, sample_df):
        custom_path = "https://example.com/custom.js"
        html = frame_display(sample_df, jspath=custom_path, return_html=True)
        assert custom_path in html

    @pytest.mark.parametrize("embed_style", ["css_only", "all"])
    def test_embed_styles(self, sample_df, embed_style, mocker):
        mock_files = mocker.patch("importlib.resources.files")
        mock_file = mocker.MagicMock()
        mock_file.read_text.return_value = "/* mock content */"
        mock_files.return_value.__truediv__.return_value = mock_file

        html = frame_display(sample_df, embed_style=embed_style, return_html=True)
        assert "/* mock content */" in html

    def test_invalid_embed_style_raises_error(self, sample_df):
        with pytest.raises(ValueError, match="Invalid value for `embed_style`"):
            frame_display(sample_df, embed_style="invalid")


class TestPandasIntegration:
    """Pandas DataFrame._repr_html_ integration."""

    def test_integration_modifies_dataframe_repr(self, sample_df):
        original = getattr(pd.DataFrame, "_repr_html_", None)

        try:
            integrate_with_pandas()
            assert hasattr(pd.DataFrame, "_repr_html_")

            html = sample_df._repr_html_()
            assert isinstance(html, str)
            assert "frame-display-table" in html

        finally:
            if original is not None:
                pd.DataFrame._repr_html_ = original
            else:
                delattr(pd.DataFrame, "_repr_html_")

    def test_integration_passes_kwargs(self, sample_df, mocker):
        original = getattr(pd.DataFrame, "_repr_html_", None)

        try:
            mock_frame_display = mocker.patch("framedisplay.display.frame_display")
            mock_frame_display.return_value = "<html>test</html>"

            integrate_with_pandas(embed_style="css_only")
            sample_df._repr_html_()

            args, kwargs = mock_frame_display.call_args
            assert kwargs["embed_style"] == "css_only"
            assert kwargs["return_html"] is True

        finally:
            if original is not None:
                pd.DataFrame._repr_html_ = original
            else:
                delattr(pd.DataFrame, "_repr_html_")


class TestEdgeCases:
    """Edge cases and stress testing."""

    def test_multiindex_dataframe(self):
        arrays = [["A", "A", "B", "B"], [1, 2, 1, 2]]
        index = pd.MultiIndex.from_arrays(arrays)
        df = pd.DataFrame({"values": [10, 20, 30, 40]}, index=index)

        html = frame_display(df, return_html=True)
        assert "frame-display-table" in html  # Should not crash

    def test_extreme_numeric_values(self):
        df = pd.DataFrame({"extreme": [float("inf"), float("-inf"), 1e308, -1e-308, np.nan]})
        html = dataframe_to_html(df)
        assert "inf" in html.lower()
        assert "null-cell" in html  # for np.nan

    def test_large_dataframe_performance(self):
        df = pd.DataFrame(
            {"col1": range(1000), "col2": ["text"] * 1000, "col3": np.random.random(1000)}
        )

        html = frame_display(df, return_html=True)
        assert "frame-display-table" in html
        assert len(html) > 1000

    def test_wide_dataframe(self):
        data = {f"col_{i}": [1, 2, 3] for i in range(50)}
        df = pd.DataFrame(data)

        html = frame_display(df, return_html=True)
        soup = BeautifulSoup(html, "html.parser")
        headers = soup.find("thead").find_all("th")
        assert len(headers) == 51  # 50 columns + index

    def test_single_row_dataframe(self):
        df = pd.DataFrame({"A": [1], "B": ["test"]})
        html = dataframe_to_html(df)
        soup = BeautifulSoup(html, "html.parser")

        rows = soup.find("tbody").find_all("tr")
        assert len(rows) == 1
        cells = rows[0].find_all(["th", "td"])
        assert len(cells) == 3  # Index + 2 data columns

    def test_index_handling(self):
        df = pd.DataFrame({"A": [1, 2]}, index=["row1", "row2"])
        html = dataframe_to_html(df)
        soup = BeautifulSoup(html, "html.parser")

        rows = soup.find("tbody").find_all("tr")
        index_cells = [row.find("th") for row in rows]
        assert index_cells[0].text == "row1"
        assert index_cells[1].text == "row2"

    def test_very_long_strings(self):
        long_string = "x" * 10000
        df = pd.DataFrame({"long": [long_string]})
        html = frame_display(df, return_html=True)
        assert "frame-display-table" in html
        assert long_string in html

    def test_categorical_dtype_detection(self):
        df = pd.DataFrame({"cat": pd.Categorical(["A", "B", "C"])})
        # Only test if dtype preserved as categorical
        if isinstance(df["cat"].dtype, pd.CategoricalDtype):
            html = frame_display(df, return_html=True)
            assert "data-dtype=category" in html

    def test_time_series_dataframe(self):
        dates = pd.date_range("2023-01-01", periods=5)
        df = pd.DataFrame(
            {
                "date": dates,
                "value": [1.1, 2.2, 3.3, 4.4, 5.5],
                "category": ["A", "B", "A", "B", "C"],
            }
        )

        html = frame_display(df, return_html=True)
        assert "data-dtype=datetime" in html
        assert "data-dtype=float" in html
        assert "data-dtype=string" in html

    def test_unicode_data(self):
        df = pd.DataFrame({"unicode": ["🚀", "测试", "🇺🇸", "∑∆∫", "café"]})

        html = frame_display(df, return_html=True)
        assert "🚀" in html
        assert "测试" in html
        assert "café" in html


class TestResourceHandling:
    """Resource loading and error handling."""

    def test_missing_js_file_fails_gracefully(self, sample_df, mocker):
        mocker.patch("importlib.resources.files", side_effect=FileNotFoundError())

        with pytest.raises(FileNotFoundError):
            frame_display(sample_df, embed_style="all")

    def test_cdn_url_format_is_valid(self):
        from framedisplay.display import JS_CDN_URL

        assert JS_CDN_URL.startswith("https://cdn.jsdelivr.net/")
        assert ".min.js" in JS_CDN_URL


class TestInitialization:
    """JavaScript initialization functionality."""

    def test_initialize_loads_and_displays_js(self, mocker):
        mock_display = mocker.patch("framedisplay.display.display")
        mock_open = mocker.patch("builtins.open")
        mock_open.return_value.__enter__.return_value.read.return_value = "test_js_content"

        initialize()

        mock_display.assert_called_once()
        html_obj = mock_display.call_args[0][0]
        assert "test_js_content" in html_obj.data
        assert '<script type="text/javascript">' in html_obj.data


class TestPropertyBased:
    """Property-based testing with hypothesis."""

    @given(st.lists(st.text(), min_size=1, max_size=10))
    def test_arbitrary_text_data_safety(self, strings):
        df = pd.DataFrame({"col": strings})
        html = dataframe_to_html(df)

        # Should never contain unescaped script tags
        assert "<script" not in html.lower()
        assert "javascript:" not in html.lower()
        assert "frame-display-table" in html

    @given(
        st.dictionaries(
            keys=st.text(min_size=1, max_size=10).filter(lambda x: x.strip()),
            values=st.lists(
                st.one_of(st.integers(), st.floats(allow_nan=True), st.text(), st.none()),
                min_size=1,
                max_size=20,
            ),
            min_size=1,
            max_size=5,
        )
    )
    def test_arbitrary_dataframes_dont_crash(self, df_data):
        try:
            df = pd.DataFrame(df_data)
            html = dataframe_to_html(df)
            assert isinstance(html, str)
            assert len(html) > 10
            # Should be valid HTML
            soup = BeautifulSoup(html, "html.parser")
            assert soup.find("table") is not None
        except (ValueError, TypeError):
            # Some combinations might create invalid DataFrames
            pass


if __name__ == "__main__":
    pytest.main([__file__, "-v"])



import re
from pathlib import Path

version_file = Path(__file__).parent / "js/src/version.js"
__version__ = re.search(
    r"const\s+version\s*=\s*['\"]([^'\"]+)['\"]", version_file.read_text()
).group(1)



import random

import numpy as np
import pandas as pd


def generate_test_dataframe(rows=1000):
    # Set random seed for reproducibility
    random.seed(42)
    np.random.seed(42)

    # Define sample data for categories
    sample_categories = [
        "Electronics",
        "Clothing",
        "Food",
        "Books",
        "Home",
        "Beauty",
        "Sports",
        "Toys",
    ]
    sample_colors = [
        "Red",
        "Blue",
        "Green",
        "Yellow",
        "Black",
        "White",
        "Purple",
        "Orange",
        "Pink",
    ]
    sample_names = [
        "Alpha",
        "Beta",
        "Gamma",
        "Delta",
        "Epsilon",
        "Zeta",
        "Eta",
        "Theta",
        "Iota",
        "Kappa",
    ]
    sample_users = ["user1", "user2", "user3", "admin1", "admin2", "guest"]

    # Create lists for each column
    data = {
        "ID": [i + 1000 for i in range(rows)],
        "Name": [
            (random.choice(sample_names) + "-" + str(i) if random.random() > 0.05 else None)
            for i in range(rows)
        ],
        "Date": [
            (
                pd.Timestamp(
                    f"{2020 + random.randint(0, 4)}-{random.randint(1, 12):02d}-{random.randint(1, 28):02d}"
                )
                if random.random() > 0.05
                else None
            )
            for _ in range(rows)
        ],
        "Price": [
            (round(random.uniform(10, 1000), 2) if random.random() > 0.05 else None)
            for _ in range(rows)
        ],
        "Quantity": [
            random.randint(1, 100) if random.random() > 0.05 else None for _ in range(rows)
        ],
        "Rating": [
            round(random.uniform(1, 5), 1) if random.random() > 0.05 else None for _ in range(rows)
        ],
        "Description": [
            (
                f"This is a sample description for item {i}. It contains details about the product."
                if random.random() > 0.05
                else None
            )
            for i in range(rows)
        ],
        "Category": [
            random.choice(sample_categories) if random.random() > 0.05 else None
            for _ in range(rows)
        ],
        "Color": [
            random.choice(sample_colors) if random.random() > 0.05 else None for _ in range(rows)
        ],
        "IsAvailable": [
            random.random() > 0.3 if random.random() > 0.05 else None for _ in range(rows)
        ],
        "CreatedBy": [
            random.choice(sample_users) if random.random() > 0.05 else None for _ in range(rows)
        ],
        "Percentage": [
            round(random.uniform(0, 100), 2) if random.random() > 0.05 else None
            for _ in range(rows)
        ],
        "Temperature": [
            (round(random.uniform(60, 100), 1) if random.random() > 0.05 else None)
            for _ in range(rows)
        ],
        "Weight": [
            (round(random.uniform(0.1, 20), 2) if random.random() > 0.05 else None)
            for _ in range(rows)
        ],
        "Height": [
            round(random.uniform(1, 3), 2) if random.random() > 0.05 else None for _ in range(rows)
        ],
        "Age": [random.randint(18, 88) if random.random() > 0.05 else None for _ in range(rows)],
        "Code": [
            (f"CODE-{random.randint(0, 9999):04d}" if random.random() > 0.05 else None)
            for _ in range(rows)
        ],
        "URL": [
            (f"https://example.com/product/{i}" if random.random() > 0.05 else None)
            for i in range(rows)
        ],
        "Email": [f"user{i}@example.com" if random.random() > 0.05 else None for i in range(rows)],
        "PhoneNumber": [
            (
                f"({random.randint(100, 999)})-{random.randint(100, 999)}-{random.randint(1000, 9999)}"
                if random.random() > 0.05
                else None
            )
            for _ in range(rows)
        ],
    }

    # Convert to pandas DataFrame
    df = pd.DataFrame(data)

    # Add some datetime columns in different formats
    df["UpdatedDate"] = pd.to_datetime(df["Date"]) + pd.to_timedelta(
        np.random.randint(1, 365, size=rows), unit="d"
    )
    df["CreatedDateTime"] = pd.to_datetime(df["Date"]) - pd.to_timedelta(
        np.random.randint(1, 30, size=rows), unit="d"
    )

    # Add a currency column formatted with $ sign
    df["FormattedPrice"] = df["Price"].apply(lambda x: f"${x:.2f}" if pd.notnull(x) else None)

    # Add a column with very long text for testing truncation
    df["LongDescription"] = df["Description"].apply(
        lambda x: (
            x + " " + "This is additional text to make the description much longer. " * 10
            if pd.notnull(x)
            else None
        )
    )

    # Add some calculated columns
    df["TotalValue"] = df["Price"] * df["Quantity"]
    df["DiscountPrice"] = df["Price"] * 0.9

    # Add a column with JSON-like strings
    df["Metadata"] = df.apply(
        lambda row: (
            f'{{"id": {row["ID"]}, "category": "{row["Category"]}", "rating": {row["Rating"]}}}'
            if pd.notnull(row["Category"]) and pd.notnull(row["Rating"])
            else None
        ),
        axis=1,
    )

    return df
